from django.db import models
from django.urls import path,reverse_lazy
from django.contrib.auth.decorators import login_required

from automatic_crud.utils import get_model
from automatic_crud.data_types import *
from automatic_crud.base_report import GetExcelReport
from automatic_crud.views_crud import *
from automatic_crud.views_crud_ajax import *

class BaseModel(models.Model):
    """Model definition for BaseModel."""

    # TODO: Define fields here
    id = models.AutoField(primary_key = True)
    model_state = models.BooleanField(default = True)
    date_created = models.DateTimeField('Fecha de Creación', auto_now=False, auto_now_add=True)
    date_modified = models.DateTimeField('Fecha de Modificación', auto_now=True, auto_now_add=False)
    date_deleted = models.DateTimeField('Fecha de Eliminación', auto_now=True, auto_now_add=False)  
    
    create_form = None
    update_form = None
    all_cruds_types = True    
    ajax_crud = False
    normal_cruds = False
    
    server_side = False
    exclude_model = False
    normal_pagination = False
    values_for_page = 10
    
    login_required = False
    permission_required = ()
    model_permissions = False
    default_permissions = False
    
    exclude_fields = ['date_created','date_modified','date_deleted','model_state']

    success_create_message = "registrado correctamente!"
    success_update_message = "actualizado correctamente!"
    success_delete_message = "eliminado correctamente!"

    error_create_message = "no se ha podido registrar!"
    error_update_message = "no se ha podido actualizar!"
    non_found_message = "No se ha encontrado un registro con estos datos!"

    create_template = None
    update_template = None
    list_template = None
    detail_template = None

    class Meta:
        """Meta definition for BaseModel."""
        abstract = True

    def get_create_form(self,form = None):
        if form != None:
            self.create_form = form
        return self.create_form

    def get_update_form(self,form = None):
        if form != None:
            self.update_form = form
        return self.update_form

    def build_message(self,message:str,custom_message = False):
        if custom_message:
            return "{0}".format(message)
        return "{0} {1}".format(self._meta.verbose_name,message)

    def get_create_url(self):
        return "{0}/create/".format(self._meta.object_name.lower())
    
    def get_list_url(self):
        return "{0}/list/".format(self._meta.object_name.lower())
    
    def get_direct_delete_url(self):
        return "{0}/direct-delete/<int:pk>/".format(self._meta.object_name.lower())

    def get_logic_delete_url(self):
        return "{0}/logic-delete/<int:pk>/".format(self._meta.object_name.lower())

    def get_update_url(self):
        return "{0}/update/<int:pk>/".format(self._meta.object_name.lower())
    
    def get_detail_url(self):
        return "{0}/detail/<int:pk>/".format(self._meta.object_name.lower())
    
    def get_excel_report_url(self):
        return "{0}/excel-report/".format(self._meta.object_name.lower())
    
    def get_alias_create_url(self):
        return "{0}-{1}-create".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_list_url(self):
        return "{0}-{1}-list".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_logic_delete_url(self):
        return "{0}-{1}-logic-delete".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_direct_delete_url(self):
        return "{0}-{1}-direct-delete".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_update_url(self):
        return "{0}-{1}-update".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_detail_url(self):
        return "{0}-{1}-detail".format(self._meta.app_label,self._meta.object_name.lower())

    def get_alias_excel_report_url(self):
        return "{0}-{1}-excel-report".format(self._meta.app_label,self._meta.object_name.lower())

    def build_generics_urls_crud(self) -> URLList:
        
        __app_name = self._meta.app_label
        __model_name = self._meta.object_name
        __model = get_model(__app_name,__model_name)
        __create_form = self.get_create_form()
        __update_form = self.get_update_form()

        urlpatterns = [
            path(
                "{0}/{1}".format(__app_name,self.get_list_url()),
                BaseList.as_view(
                    template_name = __model.list_template,
                    model = __model
                ),
                name = self.get_alias_list_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_create_url()),
                BaseCreate.as_view(
                    template_name = __model.create_template,model = __model,
                    form_class = __create_form,success_url = reverse_lazy("{0}".format(self.get_alias_list_url()))
                ),
                name = self.get_alias_create_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_detail_url()),
                BaseDetail.as_view(model = __model),
                name = self.get_alias_detail_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_update_url()),
                BaseUpdate.as_view(
                    template_name = __model.update_template,model = __model,
                    form_class = __update_form,success_url = reverse_lazy("{0}".format(self.get_alias_list_url()))
                ),
                name = self.get_alias_update_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_logic_delete_url()),
                BaseLogicDelete.as_view(
                    model = __model,
                    success_url = reverse_lazy("{0}".format(self.get_alias_list_url()))
                ),
                name = self.get_alias_logic_delete_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_direct_delete_url()),
                BaseDirectDelete.as_view(
                    model = __model,
                    success_url = reverse_lazy("{0}".format(self.get_alias_list_url()))
                ),
                name = self.get_alias_direct_delete_url()
            ),
            path(
                "{0}/{1}".format(__app_name,self.get_excel_report_url()),
                GetExcelReport.as_view(),{'__app_name':__app_name,'__model_name':__model_name},
                name = self.get_alias_excel_report_url()
            ),
        ]

        return urlpatterns

    def build_generics_urls_ajax_crud(self) -> URLList:
        __app_name = self._meta.app_label
        __model_name = self._meta.object_name
        __model = get_model(__app_name,__model_name)
        __model_context = {
            'model':__model
        }
        __model_create_form_context = {
            'model':__model,
            'form':self.get_create_form()
        }
        __model_update_form_context = {
            'model':__model,
            'form':self.get_update_form()
        }

        urlpatterns = [
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_list_url()),
                BaseListAJAX.as_view(),__model_context,
                name = "{0}-ajax".format(self.get_alias_list_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_create_url()),
                BaseCreateAJAX.as_view(),__model_create_form_context,
                name = "{0}-ajax".format(self.get_alias_create_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_detail_url()),
                BaseDetailAJAX.as_view(),__model_context,
                name = "{0}-ajax".format(self.get_alias_detail_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_update_url()),
                BaseUpdateAJAX.as_view(),__model_update_form_context,
                name = "{0}-ajax".format(self.get_alias_update_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_logic_delete_url()),
                BaseLogicDeleteAJAX.as_view(),__model_context,
                name = "{0}-ajax".format(self.get_alias_logic_delete_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_direct_delete_url()),
                BaseDirectDeleteAJAX.as_view(),__model_context,
                name = "{0}-ajax".format(self.get_alias_direct_delete_url())
            ),
            path(
                "ajax-{0}/{1}".format(__app_name,self.get_excel_report_url()),
                GetExcelReport.as_view(),{'_app_name':__app_name,'_model_name':__model_name},
                name = "{0}-ajax".format(self.get_alias_excel_report_url())
            ),
        ]

        return urlpatterns



from django.http import JsonResponse as JR

from automatic_crud.data_types import Instance,JsonResponse,DjangoForm

def jr_response(message:str,error: str,statud_code: int) -> JsonResponse:
    response = JR({'message':message,'error':error})
    response.status_code = statud_code
    return response

def success_create_message(model: Instance) -> JsonResponse:
    message = model().build_message(model.success_create_message)
    error = 'Ninguno'
    return jr_response(message,error,201)

def error_create_message(model: Instance, form:DjangoForm) -> JsonResponse:
    message = model().build_message(model.error_create_message)
    error = form.errors
    return jr_response(message,error,400)

def success_update_message(model: Instance) -> JsonResponse:
    message = model().build_message(model.success_update_message)
    error = 'Ninguno'
    return jr_response(message,error,200)

def error_update_message(model: Instance, form:DjangoForm) -> JsonResponse:
    message = model().build_message(model.error_update_message)
    error = form.errors
    return jr_response(message,error,400)

def success_delete_message(model: Instance) -> JsonResponse:
    message = model().build_message(model.success_delete_message)
    error = 'Ninguno'
    return jr_response(message,error,200)

def not_found_message(model: Instance) -> JsonResponse:
    response = JR({'error':model.non_found_message})
    response.status_code = 400
    return response


from typing import Dict,List

from django.apps import apps
from django.forms import models

from automatic_crud.data_types import Instance,DjangoForm

def get_model(__app_name:str,__model_name:str) -> Instance:
    # return the model corresponding to the application name and model name sent
    return apps.get_model(app_label = __app_name,model_name = __model_name)

def get_object(model: Instance,pk: int):
    # return the record for a pk sended
    instance = model.objects.filter(id = pk,model_state = True).first()
    if instance:
        return instance
    return None

def get_model_fields_names(__model: Instance) -> List:
    # return a list of field names from a model
    return [name for name,_ in models.fields_for_model(__model).items()]

def get_queryset(__model:Instance) -> Dict:
    # returns all records in a dictionary for a model
    return __model.objects.all().values()

def get_form(form: DjangoForm,model: Instance) -> DjangoForm:
    """
    Return a Django Form for a model, also a Django Form can be indicated
    by default the Django Form will exclude the 'state' field from the model

    """


    if form is not None:
        return models.modelform_factory(model = model,form = form)
    else:
        return models.modelform_factory(model = model,exclude = ('model_state',))

def build_template_name(template_name: str,model: Instance,action:str) -> str:
    """
    Build template name with app label from model, model name and action(list,create,update,detail)

    """


    if template_name == None:
        template_name = '{0}/{1}_{2}.html'.format(
                                        model._meta.app_label,
                                        model._meta.object_name.lower(),
                                        action
                                    )
    return template_name


from django.http import JsonResponse as JSR
from django.contrib.auth.mixins import AccessMixin
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from django.views.generic import View

class BaseCrudMixin(AccessMixin):
    model = None
    data = None
    permission_required = ()

    def get_permission_required(self):
        """
        Override this method to override the permission_required attribute.
        Must return an iterable.
        """
        if self.permission_required is None:
            raise ImproperlyConfigured(
                '{0} is missing the permission_required attribute. Define {0}.permission_required, or override '
                '{0}.get_permission_required().'.format(self.__class__.__name__)
            )
        if isinstance(self.permission_required, str):
            perms = (self.permission_required,)
        else:
            perms = self.permission_required
        return perms

    def has_permission(self):
        """
        Override this method to customize the way permissions are checked.
        """
        perms = self.get_permission_required()
        return self.request.user.has_perms(perms)

    def set_permissions(self):
        """
        Set permission_required with default permissions if default_permissions = True
        """

        if self.model.default_permissions:
            __default_permissions = ('add_{0}'.format(self.model.__name__.lower()),
                                    'view_{0}'.format(self.model.__name__.lower()),
                                    'delete_{0}'.format(self.model.__name__.lower()),
                                    'change_{0}'.format(self.model.__name__.lower()))
            self.permission_required = __default_permissions
        else:
            self.permission_required = self.model.permission_required

    def validate_permissions(self,*args, **kwargs):
        """
        Validate permissions required if model_permissions = True
        """

        if self.model.model_permissions:
            self.set_permissions()
            if not self.has_permission() and not self.request.user.is_superuser:
                response = JSR({'error': 'No tiene los permisos para realizar esta acción.'})
                response.status_code = 403
                return True,response
        return False,None

    def validate_login_required(self, *args, **kwargs):
        """
        Validate login required if login_required = True
        """
        
        if self.model.login_required:          
            if not self.request.user.is_authenticated:
                response = JSR({'error': 'No ha iniciado sesión.'})
                response.status_code = 403
                return True,response
        return False,None


class BaseCrud(BaseCrudMixin,View):

    def get_fields_for_model(self):
        """
        Return fields for model excluding exclude_fields of model
        """
        fields = [field.name for field in self.model._meta.get_fields()]
        for field in self.model.exclude_fields:            
            if field in fields:
                fields.remove(field)
        return fields


import json
import ast

from django.shortcuts import render
from django.http import HttpResponse,JsonResponse as JSR
from django.core.serializers import serialize
from django.views.generic import View

from automatic_crud.generics import BaseCrud
from automatic_crud.utils import get_object,get_form
from automatic_crud.response_messages import *

class BaseListAJAX(BaseCrud):

    def get_queryset(self):
        return self.model.objects.filter(model_state = True)

    def get_server_side_queryset(self):
        """
        Returns the values as a dictionary ordered by the order_by attribute, 
        by default order_by = id

        """

        return self.model.objects.filter(model_state = True).values().order_by(
                                                f"{self.request.GET.get('order_by','id')}")

    def server_side(self):
        """
        Returns the paged query from the server excluding the fields that have been defined in 
        self.model.exclude_fields.

        The follow attributes must be sent in request.GET:
            start: element number where the page starts
            end: element number where the page ends

        The response structure is:

            {
                'length': # amount of records,
                'objects': # list of records
            }

        For more information see: https://www.youtube.com/watch?v=89Ur7GCyLxI

        """


        start = int(self.request.GET.get('start','0'))
        end = int(self.request.GET.get('end','10'))

        object_list = []
        
        for index,instance in enumerate(self.get_server_side_queryset()[start:start+end],start):
            for exclude_field in self.model.exclude_fields:
                del instance[f'{exclude_field}']
            
            instance['index'] = index + 1
            object_list.append(instance)   
        
        self.data = {
            'length': self.get_server_side_queryset().count(),
            'objects':object_list
        }
        self.data = json.dumps(self.data)

    def normalize_data(self):
        """
        Generate an HttpResponse instance to get the serialized query and 
        delete the ['model'] key from the dictionary and convert the dictionary 
        to json and save on self.data

        """

        temp_response = JSR({'data':self.data})
        temp_data = temp_response.content.decode("UTF-8")
        temp_data = ast.literal_eval(temp_data)
        temp_data = json.loads(temp_data['data'])
        for item in temp_data:  
            del item['model']
        self.data = json.dumps(temp_data)

    def get(self, request,model,*args,**kwargs):
        """
        Return data of model

        If self.model.server_side == True return Paginated Data
        else return No Paginated Data

        """


        self.model = model

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response

        if self.model.server_side:
            self.server_side()
        else:            
            self.data = serialize('json',self.get_queryset(),
                                    fields = self.get_fields_for_model(),
                                    use_natural_foreign_keys = True)
            self.normalize_data()
        return HttpResponse(self.data, content_type="application/json")

class BaseCreateAJAX(BaseCrud):
    model = None
    form_class = None

    def post(self,request,model,form = None,*args,**kwargs):
        self.model = model

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response
        
        self.form_class = get_form(form,self.model)
        form = self.form_class(request.POST,request.FILES)        
        if form.is_valid():
            form.save()
            return success_create_message(self.model)
        return error_create_message(self.model,form)

class BaseDetailAJAX(BaseCrud):

    def get(self,request,model,*args,**kwargs):
        self.model = model

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response
        
        instance = get_object(self.model,self.kwargs['pk'])        
        if instance is not None:
            self.data = serialize(
                            'json',[instance,],
                            fields = self.get_fields_for_model(),
                            use_natural_foreign_keys = True,
                            use_natural_primary_keys = True
                        )
            return HttpResponse(self.data, content_type="application/json")
        return not_found_message(self.model)

class BaseUpdateAJAX(BaseCrud):
    model = None
    form_class = None

    def post(self,request,model,form = None,*args,**kwargs):
        self.model = model        

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response
     
        self.form_class = get_form(form,self.model)        
        instance = get_object(self.model,self.kwargs['pk'])        
        if instance is not None:
            form = self.form_class(request.POST,request.FILES,instance = instance)            
            if form.is_valid():
                form.save()
                return success_update_message(self.model)        
            else:
                return error_update_message(self.model,form)
        return not_found_message(self.model)

class BaseDirectDeleteAJAX(BaseCrud):
    model = None

    def delete(self,request,model,*args,**kwargs):
        self.model = model

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response

        instance = get_object(self.model,self.kwargs['pk'])        
        if instance is not None:
            instance.delete()
            return success_delete_message(self.model)
        return not_found_message(self.model)

class BaseLogicDeleteAJAX(BaseCrud):
    model = None

    def delete(self,request,model,*args,**kwargs):
        self.model = model

        # login required validation
        validation_login_required,response = self.validate_login_required()
        if validation_login_required:
            return response
        
        # permission required validation
        validation_permissions,response = self.validate_permissions()
        if validation_permissions:
            return response

        instance = get_object(self.model,self.kwargs['pk'])        
        if instance is not None:
            self.model.objects.filter(id = self.kwargs['pk']).update(model_state = False)
            return success_delete_message(self.model)
        return not_found_message(self.model)


from django.apps import apps

from automatic_crud.models import BaseModel

def register_models():
    """
    Register models with automatic cruds excluding models with exclude_model = True
    Return urlspatterns with automatic cruds
    """

    urlpatterns = []
    exclude_models = ['ContentType','LogEntry','Session','Permission','Group']
    models = apps.get_models()
    
    for model in models:

        if isinstance(model(),BaseModel):
            try:
                if model.__name__ not in exclude_models:
                    
                    if not model.exclude_model:

                        if model.all_cruds_types:
                            urlpatterns += model().build_generics_urls_crud()
                            urlpatterns += model().build_generics_urls_ajax_crud()
                        
                        else:
                            
                            if model.ajax_crud:         
                                urlpatterns += model().build_generics_urls_ajax_crud()
                            if model.normal_cruds:
                                urlpatterns += model().build_generics_urls_crud()                            

            except:
                pass

    return urlpatterns